#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External Libraries
import numpy as np
import pandas as pd
from sklearn.metrics import roc_curve, roc_auc_score

#-----------------------------------------------------#
#         Computation: Classification Metrics         #
#-----------------------------------------------------#
def compute_metrics(preds, labels, n_labels, threshold=None):
    """ Function for computing various classification metrics.

    !!! info "Computed Metrics"
        F1, Accuracy, Sensitivity, Specificity, AUROC (AUC), Precision, FPR, FNR,
        FDR, TruePositives, TrueNegatives, FalsePositives, FalseNegatives

    Args:
        preds (numpy.ndarray):          A NumPy array of predictions formatted with shape (n_samples, n_labels). Provided by
                                        [NeuralNetwork][aucmedi.neural_network.model].
        labels (numpy.ndarray):         Classification list with One-Hot Encoding. Provided by
                                        [input_interface][aucmedi.data_processing.io_data.input_interface].
        n_labels (int):                 Number of classes. Provided by [input_interface][aucmedi.data_processing.io_data.input_interface].
        threshold (float):              Only required for multi_label data. Threshold value if prediction is positive.

    Returns:
        metrics (pandas.DataFrame):     Dataframe containing all computed metrics (except ROC).
    """
    df_list = []
    for c in range(0, n_labels):
        # Initialize variables
        data_dict = {}

        # Identify truth and prediction for class c
        truth = labels[:, c]
        if threshold is None:
            pred_argmax = np.argmax(preds, axis=-1)
            pred = (pred_argmax == c).astype(np.int8)
        else:
            pred = np.where(preds[:, c] >= threshold, 1, 0)
        # Obtain prediction confidence (probability)
        pred_prob = preds[:, c]

        # Compute the confusion matrix
        tp, tn, fp, fn = compute_CM(truth, pred)
        data_dict["TP"] = tp
        data_dict["TN"] = tn
        data_dict["FP"] = fp
        data_dict["FN"] = fn

        # Compute several metrics based on confusion matrix
        data_dict["Sensitivity"] = np.divide(tp, tp+fn)
        data_dict["Specificity"] = np.divide(tn, tn+fp)
        data_dict["Precision"] = np.divide(tp, tp+fp)
        data_dict["FPR"] = np.divide(fp, fp+tn)
        data_dict["FNR"] = np.divide(fn, fn+tp)
        data_dict["FDR"] = np.divide(fp, fp+tp)
        data_dict["Accuracy"] = np.divide(tp+tn, tp+tn+fp+fn)
        data_dict["F1"] = np.divide(2*tp, 2*tp+fp+fn)

        # Compute area under the ROC curve
        try:
            data_dict["AUC"] = roc_auc_score(truth, pred_prob)
        except:
            print("ROC AUC score is not defined.")

        # Parse metrics to dataframe
        df = pd.DataFrame.from_dict(data_dict, orient="index",
                                    columns=["score"])
        df = df.reset_index()
        df.rename(columns={"index": "metric"}, inplace=True)
        df["class"] = c

        # Append dataframe to list
        df_list.append(df)

    # Combine dataframes
    df_final = pd.concat(df_list, axis=0, ignore_index=True)
    # Return final dataframe
    return df_final

#-----------------------------------------------------#
#            Computation: Confusion Matrix            #
#-----------------------------------------------------#
def compute_confusion_matrix(preds, labels, n_labels):
    """ Function for computing a confusion matrix.

    Args:
        preds (numpy.ndarray):          A NumPy array of predictions formatted with shape (n_samples, n_labels). Provided by
                                        [NeuralNetwork][aucmedi.neural_network.model].
        labels (numpy.ndarray):         Classification list with One-Hot Encoding. Provided by
                                        [input_interface][aucmedi.data_processing.io_data.input_interface].
        n_labels (int):                 Number of classes. Provided by [input_interface][aucmedi.data_processing.io_data.input_interface].

    Returns:
        rawcm (numpy.ndarray):          NumPy matrix with shape (n_labels, n_labels).
    """
    preds_argmax = np.argmax(preds, axis=-1)
    labels_argmax = np.argmax(labels, axis=-1)
    rawcm = np.zeros((n_labels, n_labels))
    for i in range(0, labels.shape[0]):
        rawcm[labels_argmax[i]][preds_argmax[i]] += 1
    return rawcm

#-----------------------------------------------------#
#             Computation: ROC Coordinates            #
#-----------------------------------------------------#
def compute_roc(preds, labels, n_labels):
    """ Function for computing the data data of a ROC curve (FPR and TPR).

    Args:
        preds (numpy.ndarray):          A NumPy array of predictions formatted with shape (n_samples, n_labels). Provided by
                                        [NeuralNetwork][aucmedi.neural_network.model].
        labels (numpy.ndarray):         Classification list with One-Hot Encoding. Provided by
                                        [input_interface][aucmedi.data_processing.io_data.input_interface].
        n_labels (int):                 Number of classes. Provided by [input_interface][aucmedi.data_processing.io_data.input_interface].
    Returns:
        fpr_list (list of list):        List containing a list of false positive rate points for each class. Shape: (n_labels, tpr_coords).
        tpr_list (list of list):        List containing a list of true positive rate points for each class. Shape: (n_labels, fpr_coords).
    """
    fpr_list = []
    tpr_list = []
    for i in range(0, n_labels):
        truth_class = labels[:, i].astype(int)
        pdprob_class = preds[:, i]
        fpr, tpr, _ = roc_curve(truth_class, pdprob_class)
        fpr_list.append(fpr)
        tpr_list.append(tpr)
    return fpr_list, tpr_list

#-----------------------------------------------------#
#                     Subroutines                     #
#-----------------------------------------------------#
# Compute confusion matrix
def compute_CM(gt, pd):
    tp = 0
    tn = 0
    fp = 0
    fn = 0
    for i in range(0, len(gt)):
        if gt[i] == 1 and pd[i] == 1 : tp += 1
        elif gt[i] == 1 and pd[i] == 0 : fn += 1
        elif gt[i] == 0 and pd[i] == 0 : tn += 1
        elif gt[i] == 0 and pd[i] == 1 : fp += 1
        else : print("ERROR at confusion matrix", i)
    return tp, tn, fp, fn



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from abc import ABC, abstractmethod

#-----------------------------------------------------#
#         Abstract Base Class for Subfunctions        #
#-----------------------------------------------------#
class Subfunction_Base(ABC):
    """ An abstract base class for a Subfunction class.

    A child of this ABC can be used as a [Subfunction][aucmedi.data_processing.subfunctions]
    and be passed to a [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    This class provides functionality for running the transform function,
    which preprocesses an image during the processing (batch preparation) of the DataGenerator.

    ???+ example "Create a custom Subfunction"
        ```python
        from aucmedi.data_processing.subfunctions.sf_base import Subfunction_Base

        class My_custom_Subfunction(Subfunction_Base):
            def __init__(self):                 # you can pass here class variables
                pass

            def transform(self, image):
                new_image = image + 1.0         # do some operation
                return new_image                # return modified image
        ```

    ???+ info "Required Functions"
        | Function            | Description                                |
        | ------------------- | ------------------------------------------ |
        | `__init__()`        | Object creation function.                  |
        | `transform()`       | Transform the image.                       |

    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    @abstractmethod
    def __init__(self):
        """ Functions which will be called during the Subfunction object creation.

        This function can be used to pass variables and options in the Subfunction instance.
        The are no mandatory required parameters for the initialization.
        """
        pass
    #---------------------------------------------#
    #                Transformation               #
    #---------------------------------------------#
    @abstractmethod
    def transform(self, image):
        """ Transform the image according to the subfunction during preprocessing (training + prediction).

        It is required to return the transformed image object (as NumPy array).

        Args:
            image (numpy.ndarray):      Image encoded as NumPy matrix with 1 or 3 channels. (e.g. 224x224x3)

        Returns:
            image (numpy.ndarray):      Transformed image encoded as NumPy matrix with 1 or 3 channels. (e.g. 224x224x3)
        """
        return image



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from volumentations import Compose
from volumentations import augmentations as ai
import warnings
import numpy as np
import random

#-----------------------------------------------------#
#             AUCMEDI Volume Augmentation             #
#-----------------------------------------------------#
class VolumeAugmentation():
    """ The Volume Augmentation class performs diverse augmentation methods on given
        numpy array. The class acts as an easy to use function/interface for applying
        all types of augmentations with just one function call.

    The class can be configured beforehand by selecting desired augmentation techniques
    and method ranges or strength.
    Afterwards, the class is passed to the [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator]
    which utilizes it during batch generation.

    The specific configurations of selected methods can be adjusted by class variables.

    ???+ abstract "Build on top of the library"
        Volumentations based on albumentations.

        ```
        volumentations: Originally developed by @ashawkey and @ZFTurbo.
                        Initially inspired by albumentations library for augmentation of 2D images.
                        https://github.com/muellerdo/volumentations
                        https://github.com/ZFTurbo/volumentations
                        https://github.com/ashawkey/volumentations
        albumentations: https://github.com/albumentations-team/albumentations
        ```

        The Volumentations package was further continued by us to ensure ongoing development and support.

        For more details, please read the README under:
        https://github.com/muellerdo/volumentations
    """
    #-----------------------------------------------------#
    #              Augmentation Configuration             #
    #-----------------------------------------------------#
    # Define augmentation operator
    operator = None
    # Option for augmentation refinement (padding, cropping and clipping)
    refine = True
    # Augmentation: Flip
    aug_flip = False
    aug_flip_p = 0.5
    # Augmentation: 90 degree rotate
    aug_rotate = False
    aug_rotate_p = 0.5
    # Augmentation: Brightness
    aug_brightness = False
    aug_brightness_p = 0.5
    aug_brightness_limits = 0.1
    # Augmentation: Contrast
    aug_contrast = False
    aug_contrast_p = 0.5
    aug_contrast_limits = 0.1
    # Augmentation: Saturation shift
    aug_saturation = False
    aug_saturation_p = 0.5
    aug_saturation_limits = 0.1
    # Augmentation: Hue shift
    aug_hue = False
    aug_hue_p = 0.5
    aug_hue_limits = 0.1
    # Augmentation: Scale
    aug_scale = False
    aug_scale_p = 0.5
    aug_scale_limits = (0.9, 1.1)
    # Augmentation: Crop
    aug_crop = False
    aug_crop_p = 0.5
    aug_crop_shape = (64, 64, 64)
    # Augmentation: Grid Distortion
    aug_gridDistortion = False
    aug_gridDistortion_p = 0.5
    # Augmentation: Image Compression (JPEG)
    aug_compression = False
    aug_compression_p = 0.5
    aug_compression_limits = (90, 100)
    # Augmentation: Gaussian Noise
    aug_gaussianNoise = False
    aug_gaussianNoise_p = 0.5
    # Augmentation: Gaussian Blur
    aug_gaussianBlur = False
    aug_gaussianBlur_p = 0.5
    # Augmentation: Downscale
    aug_downscaling = False
    aug_downscaling_p = 0.5
    aug_downscaling_effect = 0.25
    # Augmentation: Gamma
    aug_gamma = False
    aug_gamma_p = 0.5
    aug_gamma_limit = (90, 110)
    # Augmentation: Elastic Transformation
    aug_elasticTransform = False
    aug_elasticTransform_p = 0.5

    #-----------------------------------------------------#
    #                    Initialization                   #
    #-----------------------------------------------------#
    def __init__(self, flip=True, rotate=True, brightness=True, contrast=True,
                 saturation=True, hue=True, scale=True, crop=False,
                 grid_distortion=False, compression=False, gaussian_noise=False,
                 gaussian_blur=False, downscaling=False, gamma=False,
                 elastic_transform=False):
        """ Initialization function for the Volume Augmentation interface.

        With boolean switches, it is possible to selected desired augmentation techniques.
        Recommended augmentation configurations are defined as class variables.
        Of course, these configs can be adjusted if needed.

        Args:
            flip (bool):                    Boolean, whether flipping should be performed as data augmentation.
            rotate (bool):                  Boolean, whether rotations should be performed as data augmentation.
            brightness (bool):              Boolean, whether brightness changes should be added as data augmentation.
            contrast (bool):                Boolean, whether contrast changes should be added as data augmentation.
            saturation (bool):              Boolean, whether saturation changes should be added as data augmentation.
            hue (bool):                     Boolean, whether hue changes should be added as data augmentation.
            scale (bool):                   Boolean, whether scaling should be performed as data augmentation.
            crop (bool):                    Boolean, whether scaling cropping be performed as data augmentation.
            grid_distortion (bool):         Boolean, whether grid_distortion should be performed as data augmentation.
            compression (bool):             Boolean, whether compression should be performed as data augmentation.
            gaussian_noise (bool):          Boolean, whether gaussian noise should be added as data augmentation.
            gaussian_blur (bool):           Boolean, whether gaussian blur should be added as data augmentation.
            downscaling (bool):             Boolean, whether downscaling should be added as data augmentation.
            gamma (bool):                   Boolean, whether gamma changes should be added as data augmentation.
            elastic_transform (bool):       Boolean, whether elastic deformation should be performed as data augmentation.

        !!! warning
            If class variables (attributes) are modified, the internal augmentation operator
            has to be rebuild via the following call:

            ```python
            # initialize
            aug = VolumeAugmentation(flip=True)

            # set probability to 100% = always
            aug.aug_flip_p = 1.0
            # rebuild
            aug.build()
            ```

        Attributes:
            refine (bool):                  Boolean, whether clipping to [0,255] and padding/cropping should be performed if outside of range.
            aug_flip_p (float):             Probability of flipping application if activated. Default=0.5.
            aug_rotate_p (float):           Probability of rotation application if activated. Default=0.5.
            aug_brightness_p (float):       Probability of brightness application if activated. Default=0.5.
            aug_contrast_p (float):         Probability of contrast application if activated. Default=0.5.
            aug_saturation_p (float):       Probability of saturation application if activated. Default=0.5.
            aug_hue_p (float):              Probability of hue application if activated. Default=0.5.
            aug_scale_p (float):            Probability of scaling application if activated. Default=0.5.
            aug_crop_p (float):             Probability of crop application if activated. Default=0.5.
            aug_grid_distortion_p (float):  Probability of grid_distortion application if activated. Default=0.5.
            aug_compression_p (float):      Probability of compression application if activated. Default=0.5.
            aug_gaussianNoise_p (float):    Probability of gaussian noise application if activated. Default=0.5.
            aug_gaussianBlur_p (float):     Probability of gaussian blur application if activated. Default=0.5.
            aug_downscaling_p (float):      Probability of downscaling application if activated. Default=0.5.
            aug_gamma_p (float):            Probability of gamma application if activated. Default=0.5.
            aug_elasticTransform_p (float): Probability of elastic deformation application if activated. Default=0.5.
        """
        # Cache class variables
        self.aug_flip = flip
        self.aug_rotate = rotate
        self.aug_brightness = brightness
        self.aug_contrast = contrast
        self.aug_scale = scale
        self.aug_crop = crop
        self.aug_saturation = saturation
        self.aug_hue = hue
        self.aug_compression = compression
        self.aug_gaussianNoise = gaussian_noise
        self.aug_gaussianBlur= gaussian_blur
        self.aug_downscaling = downscaling
        self.aug_gamma = gamma
        self.aug_gridDistortion = grid_distortion
        self.aug_elasticTransform = elastic_transform
        # Build augmentation operator
        self.build()

    #-----------------------------------------------------#
    #                Albumentations Builder               #
    #-----------------------------------------------------#
    def build(self):
        """ Builds the albumenations augmentator by initializing  all transformations.

        The activated transformation and their configurations are defined as
        class variables.

        -> Builds a new self.operator
        """
        # Initialize transform list
        transforms = []
        # Fill transform list
        if self.aug_flip:
            tf = ai.Flip(p=self.aug_flip_p)
            transforms.append(tf)
        if self.aug_rotate:
            tf = ai.RandomRotate90(p=self.aug_rotate_p)
            transforms.append(tf)
        if self.aug_brightness:
            tf = ai.ColorJitter(brightness=self.aug_brightness_limits,
                                contrast=0, hue=0, saturation=0,
                                p=self.aug_brightness_p)
            transforms.append(tf)
        if self.aug_contrast:
            tf = ai.ColorJitter(brightness=0, contrast=self.aug_contrast_limits,
                                hue=0, saturation=0,
                                p=self.aug_contrast_p)
            transforms.append(tf)
        if self.aug_saturation:
            tf = ai.ColorJitter(brightness=0, contrast=0, hue=0,
                                saturation=self.aug_saturation_limits,
                                p=self.aug_saturation_p)
            transforms.append(tf)
        if self.aug_hue:
            tf = ai.ColorJitter(brightness=0, contrast=0, saturation=0,
                                hue=self.aug_hue_limits,
                                p=self.aug_hue_p)
            transforms.append(tf)
        if self.aug_scale:
            tf = ai.RandomScale(scale_limit=self.aug_scale_limits,
                                p=self.aug_scale_p)
            transforms.append(tf)
        if self.aug_crop:
            tf = ai.RandomCrop(self.aug_crop_shape,
                               p=self.aug_crop_p)
            transforms.append(tf)
        if self.aug_gridDistortion:
            tf = ai.GridDistortion(p=self.aug_gridDistortion_p)
            transforms.append(tf)
        if self.aug_compression:
            tf = ai.ImageCompression(quality_lower=self.aug_compression_limits[0],
                                     quality_upper=self.aug_compression_limits[1],
                                     p=self.aug_compression_p)
            transforms.append(tf)
        if self.aug_gaussianNoise:
            tf = ai.GaussianNoise(p=self.aug_gaussianNoise_p)
            transforms.append(tf)
        if self.aug_gaussianBlur:
            tf = ai.GlassBlur(p=self.aug_gaussianBlur_p)
            transforms.append(tf)
        if self.aug_downscaling:
            tf = ai.Downscale(scale_min=self.aug_downscaling_effect,
                              scale_max=self.aug_downscaling_effect,
                              p=self.aug_downscaling_p)
            transforms.append(tf)
        if self.aug_gamma:
            tf = ai.RandomGamma(gamma_limit=self.aug_gamma_limit,
                                p=self.aug_gamma_p)
            transforms.append(tf)
        if self.aug_elasticTransform:
            tf = ai.ElasticTransform(p=self.aug_elasticTransform_p)
            transforms.append(tf)

        # Compose transforms
        self.operator = Compose(transforms)

    #-----------------------------------------------------#
    #                 Perform Augmentation                #
    #-----------------------------------------------------#
    """ Performs image augmentation with defined configuration on an image.

    This **internal** function is called in the DataGenerator during batch generation.

    Args:
        image (numpy.ndarray):          An image encoded as NumPy array with shape (z, y, x, channels).
    Returns:
        aug_image (numpy.ndarray):      An augmented / transformed image.
    """
    def apply(self, image):
        # Verify that image is in grayscale/RGB encoding
        if np.min(image) < 0 or np.max(image) > 255:
            warnings.warn("Image Augmentation: A value of the image is lower than 0 or higher than 255.",
                          "Volumentations expects images to be in grayscale/RGB!",
                          np.min(image), np.max(image))
        # Cache image shape
        org_shape = image.shape
        # Perform image augmentation
        aug_image = self.operator(image=image)["image"]
        # Perform padding & cropping if image shape changed
        if self.refine and aug_image.shape != org_shape:
            aug_image = ai.pad(aug_image, new_shape=org_shape)
            offset = (random.random(), random.random(), random.random())
            aug_image = ai.random_crop(aug_image,
                                       org_shape[0], org_shape[1], org_shape[2],
                                       offset[0], offset[1], offset[2])
        # Perform clipping if image is out of grayscale/RGB encodings
        if self.refine and (np.min(aug_image) < 0 or np.max(aug_image) > 255):
            aug_image = np.clip(aug_image, a_min=0, a_max=255)
        # Return augmented image
        return aug_image



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from abc import ABC, abstractmethod

#-----------------------------------------------------#
#         Abstract Base Class for Aggregation         #
#-----------------------------------------------------#
class Aggregate_Base(ABC):
    """ An abstract base class for an Aggregation class.

    ```
    Assembled predictions encoded in a NumPy matrix with shape (N_models, N_classes).
    Example: [[0.5, 0.4, 0.1],
              [0.4, 0.3, 0.3],
              [0.5, 0.2, 0.3]]
    -> shape (3, 3)

    Merged prediction encoded in a NumPy matrix with shape (1, N_classes).
    Example: [[0.4, 0.3, 0.3]]
    -> shape (1, 3)
    ```

    ???+ example "Create a custom Aggregation class"
        ```python
        from aucmedi.ensemble.aggregate.agg_base import Aggregate_Base

        class My_custom_Aggregate(Aggregate_Base):
            def __init__(self):                 # you can pass class variables here
                pass

            def aggregate(self, preds):
                preds_combined = np.mean(preds, axis=0)     # do some combination operation
                return preds_combined                       # return combined predictions
        ```

    !!! info "Required Functions"
        | Function            | Description                                                |
        | ------------------- | ---------------------------------------------------------- |
        | `__init__()`        | Object creation function.                                  |
        | `aggregate()`       | Merge multiple class predictions into a single prediction. |
    """
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    @abstractmethod
    def __init__(self):
        """ Initialization function which will be called during the Aggregation object creation.

        This function can be used to pass variables and options in the Aggregation instance.
        There are no mandatory parameters for the initialization.
        """
        pass
    #---------------------------------------------#
    #                  Aggregate                  #
    #---------------------------------------------#
    @abstractmethod
    def aggregate(self, preds):
        """ Aggregate the image by merging multiple predictions into a single one.

        It is required to return the merged predictions (as NumPy matrix).
        It is possible to pass configurations through the initialization function for this class.

        Args:
            preds (numpy.ndarray):      Assembled predictions encoded in a NumPy matrix with shape (N_models, N_classes).
        Returns:
            pred (numpy.ndarray):       Merged prediction encoded in a NumPy matrix with shape (1, N_classes).
        """
        return pred



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
#Internal libraries
from aucmedi.utils.class_weights import *

#-----------------------------------------------------#
#         Unittest: Class Weight Computations         #
#-----------------------------------------------------#
class classweightTEST(unittest.TestCase):
    # Create random classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Create classification labels
        self.labels_ohe = np.zeros((25, 4), dtype=np.uint8)
        for i in range(0, 25):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1

    #-------------------------------------------------#
    #           Class weights (multi-class)           #
    #-------------------------------------------------#
    def test_classweights_multiclass(self):
        cwl, cwd = compute_class_weights(self.labels_ohe)
        self.assertEqual(len(cwl), 4)
        self.assertTrue(isinstance(cwl[0], float))
        self.assertTrue(isinstance(cwl[1], float))
        self.assertTrue(isinstance(cwl[2], float))
        self.assertTrue(isinstance(cwl[3], float))
        self.assertEqual(len(cwd), 4)
        self.assertTrue(isinstance(cwd[0], float))
        self.assertTrue(isinstance(cwd[1], float))
        self.assertTrue(isinstance(cwd[2], float))
        self.assertTrue(isinstance(cwd[3], float))

    #-------------------------------------------------#
    #           Class weights (multi-label)           #
    #-------------------------------------------------#
    def test_classweights_multilabel(self):
        class_weights = compute_multilabel_weights(self.labels_ohe)
        self.assertEqual(len(class_weights), 4)
        self.assertTrue(isinstance(class_weights[0], float))
        self.assertTrue(isinstance(class_weights[1], float))
        self.assertTrue(isinstance(class_weights[2], float))
        self.assertTrue(isinstance(class_weights[3], float))

    #-------------------------------------------------#
    #                 Sample weights                  #
    #-------------------------------------------------#
    def test_sampleweights(self):
        class_weights = compute_sample_weights(self.labels_ohe)
        self.assertEqual(len(class_weights), 25)
        self.assertTrue(isinstance(class_weights[0], float))
        self.assertTrue(isinstance(class_weights[5], float))
        self.assertTrue(isinstance(class_weights[10], float))
        self.assertTrue(isinstance(class_weights[20], float))
        self.assertTrue(isinstance(class_weights[24], float))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import numpy as np
import SimpleITK as sitk

#-----------------------------------------------------#
#              SITK Loader for AUCMEDI IO             #
#-----------------------------------------------------#
def sitk_loader(sample, path_imagedir, image_format=None, grayscale=True,
                resampling=(1.0, 1.0, 1.0), outside_value=0, **kwargs):
    """ SimpleITK Loader for loading of CT/MRI scans in NIfTI (nii) or Metafile (mha) format within the AUCMEDI pipeline.

    The SimpleITK Loader is an IO_loader function, which have to be passed to the
    [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

    This loader is intended to load only 3D volumes with annotated voxel spacings.

    By default, volumes are normalized to voxel spacing 1.0 x 1.0 x 1.0. <br>
    You can define a custom voxel spacing for the loader, by passing a tuple of spacings as parameter 'resampling'. <br>

    ???+ info
        The SimpleITK Loader utilizes SimpleITK for sample loading: <br>
        https://simpleitk.readthedocs.io/en/master/IO.html

    ???+ example
        ```python
        # Import required libraries
        from aucmedi import *
        from aucmedi.data_processing.io_loader import sitk_loader

        # Initialize input data reader
        ds = input_interface(interface="csv",
                            path_imagedir="dataset/nii_files/",
                            path_data="dataset/annotations.csv",
                            ohe=False, col_sample="ID", col_class="diagnosis")
        (samples, class_ohe, nclasses, class_names, image_format) = ds

        # Initialize DataGenerator with sitk_loader
        data_gen = DataGenerator(samples, "dataset/nii_files/", labels=class_ohe,
                                image_format=image_format, resize=None,
                                grayscale=True, resampling=(2.10, 1.48, 1.48),
                                loader=sitk_loader)
        ```

    Args:
        sample (str):               Sample name/index of an image.
        path_imagedir (str):        Path to the directory containing the images.
        image_format (str):         Image format to add at the end of the sample index for image loading.
        grayscale (bool):           Boolean, whether images are grayscale or RGB.
        resampling (tuple of float):Tuple of 3x floats with z,y,x mapping encoding voxel spacing.
                                    If passing `None`, no normalization will be performed.
        **kwargs (dict):            Additional parameters for the sample loader.
    """
    # Get image path
    if image_format : img_file = sample + "." + image_format
    else : img_file = sample
    path_img = os.path.join(path_imagedir, img_file)
    # Load image via the SimpleITK package
    sample_itk = sitk.ReadImage(path_img)
    # Perform resampling
    if resampling is not None:
        # Extract information from sample
        shape = sample_itk.GetSize()
        spacing = sample_itk.GetSpacing()
        # Reverse resampling spacing to sITK mapping (z,y,x -> x,y,z)
        new_spacing = resampling[::-1]
        # Estimate output shape after resampling
        output_shape = []
        for t in zip(shape, spacing, new_spacing):
            s = int(t[0] * t[1] / t[2])
            output_shape.append(s)
        output_shape = tuple(output_shape)
        # Perform resampling via sITK
        sample_itk_resampled = sitk.Resample(sample_itk,
                                             output_shape,
                                             sitk.Transform(),
                                             sitk.sitkLinear,
                                             sample_itk.GetOrigin(),
                                             new_spacing,
                                             sample_itk.GetDirection(),
                                             outside_value)
    # Skip resampling if None
    else : sample_itk_resampled = sample_itk
    # Convert to NumPy
    img = sitk.GetArrayFromImage(sample_itk_resampled)
    # Add single channel axis
    if len(img.shape) == 3 : img = np.expand_dims(img, axis=-1)
    # Return image
    return img



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
import tempfile
import os
#Internal libraries
from aucmedi.ensemble.metalearner import *

#-----------------------------------------------------#
#                Unittest: Metalearner                #
#-----------------------------------------------------#
class MetelearnerTEST(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Create model prediction data
        self.pred_data = np.random.rand(25, 12)
        # Create classification labels
        self.labels_ohe = np.zeros((25, 4), dtype=np.uint8)
        for i in range(0, 25):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1
        # Temporary model directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".model")

    #-------------------------------------------------#
    #               Logistic Regression               #
    #-------------------------------------------------#
    def test_LogisticRegression_create(self):
        # Initializations
        ml = LogisticRegression()
        self.assertTrue("logistic_regression" in metalearner_dict)
        ml = metalearner_dict["logistic_regression"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_LogisticRegression_usage(self):
        # Initializations
        ml = LogisticRegression()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                   Naive Bayes                   #
    #-------------------------------------------------#
    def test_NaiveBayes_create(self):
        # Initializations
        ml = NaiveBayes()
        self.assertTrue("naive_bayes" in metalearner_dict)
        ml = metalearner_dict["naive_bayes"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_NaiveBayes_usage(self):
        # Initializations
        ml = NaiveBayes()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #              Support Vector Machine             #
    #-------------------------------------------------#
    def test_SVM_create(self):
        # Initializations
        ml = SupportVectorMachine()
        self.assertTrue("support_vector_machine" in metalearner_dict)
        ml = metalearner_dict["support_vector_machine"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_SVM_usage(self):
        # Initializations
        ml = SupportVectorMachine()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                 Gaussian Process                #
    #-------------------------------------------------#
    def test_GaussianProcess_create(self):
        # Initializations
        ml = GaussianProcess()
        self.assertTrue("gaussian_process" in metalearner_dict)
        ml = metalearner_dict["gaussian_process"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_GaussianProcess_usage(self):
        # Initializations
        ml = GaussianProcess()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                  Decision Tree                  #
    #-------------------------------------------------#
    def test_DecisionTree_create(self):
        # Initializations
        ml = DecisionTree()
        self.assertTrue("decision_tree" in metalearner_dict)
        ml = metalearner_dict["decision_tree"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_DecisionTree_usage(self):
        # Initializations
        ml = DecisionTree()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                    Best Model                   #
    #-------------------------------------------------#
    def test_BestModel_create(self):
        # Initializations
        ml = BestModel()
        self.assertTrue("best_model" in metalearner_dict)
        ml = metalearner_dict["best_model"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_BestModel_usage(self):
        # Initializations
        ml = BestModel()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                  Weighted Mean                  #
    #-------------------------------------------------#
    def test_AveragingWeightedMean_create(self):
        # Initializations
        ml = AveragingWeightedMean()
        self.assertTrue("weighted_mean" in metalearner_dict)
        ml = metalearner_dict["weighted_mean"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_AveragingWeightedMean_usage(self):
        # Initializations
        ml = AveragingWeightedMean()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                  Random Forest                  #
    #-------------------------------------------------#
    def test_RandomForest_create(self):
        # Initializations
        ml = RandomForest()
        self.assertTrue("random_forest" in metalearner_dict)
        ml = metalearner_dict["random_forest"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_RandomForest_usage(self):
        # Initializations
        ml = RandomForest()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #               k-Nearest Neighbors               #
    #-------------------------------------------------#
    def test_KNearestNeighbors_create(self):
        # Initializations
        ml = KNearestNeighbors()
        self.assertTrue("k_neighbors" in metalearner_dict)
        ml = metalearner_dict["k_neighbors"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_KNearestNeighbors_usage(self):
        # Initializations
        ml = KNearestNeighbors()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))

    #-------------------------------------------------#
    #                MLP Neural Network               #
    #-------------------------------------------------#
    def test_MLP_create(self):
        # Initializations
        ml = MLP()
        self.assertTrue("mlp" in metalearner_dict)
        ml = metalearner_dict["mlp"]()
        # Storage
        model_path = os.path.join(self.tmp_data.name, "ml_model.pickle")
        self.assertFalse(os.path.exists(model_path))
        ml.dump(model_path)
        self.assertTrue(os.path.exists(model_path))
        # Loading
        ml.model = None
        self.assertTrue(ml.model is None)
        ml.load(model_path)
        self.assertFalse(ml.model is None)
        # Cleanup
        os.remove(model_path)

    def test_MLP_usage(self):
        # Initializations
        ml = MLP()
        # Training
        ml.train(x=self.pred_data, y=self.labels_ohe)
        # Inference
        preds = ml.predict(data=self.pred_data)
        # Check
        self.assertTrue(np.array_equal(preds.shape, (25,4)))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" Library of evaluation functions in the AUCMEDI environment.

The idea of the evaluation function library is to provide quick and simple functionality
for inspecting and analyzing data as well as results generated by AUCMEDI.

AUCMEDI currently supports the following evaluation techniques:

| Evaluation Technique                                     | Description                                                                           |
| -------------------------------------------------------- | ------------------------------------------------------------------------------------- |
| [Dataset Analysis][aucmedi.evaluation.dataset]           | Peak into the loaded medical imaging dataset.                                         |
| [Fitting Evaluation][aucmedi.evaluation.fitting]         | Evaluate the fitting curve of a model training process.                               |
| [Performance Evaluation][aucmedi.evaluation.performance] | Evaluate the performance of a single model / prediction list through various metrics. |
| [Performance Comparison][aucmedi.evaluation.comparison]  | Compare the performance of predictions from multiple models.                          |

"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from aucmedi.evaluation.fitting import evaluate_fitting
from aucmedi.evaluation.performance import evaluate_performance
from aucmedi.evaluation.comparison import evaluate_comparison
from aucmedi.evaluation.dataset import evaluate_dataset



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# Import XAI methods
from aucmedi.xai.methods.gradcam import GradCAM
from aucmedi.xai.methods.gradcam_pp import GradCAMpp
from aucmedi.xai.methods.saliency import SaliencyMap
from aucmedi.xai.methods.guided_backprop import GuidedBackpropagation
from aucmedi.xai.methods.integrated_gradients import IntegratedGradients
from aucmedi.xai.methods.gradcam_guided import GuidedGradCAM
from aucmedi.xai.methods.occlusion_sensitivity import OcclusionSensitivity
from aucmedi.xai.methods.lime_pro import LimePro
from aucmedi.xai.methods.lime_con import LimeCon

#-----------------------------------------------------#
#                XAI method dictionary                #
#-----------------------------------------------------#
xai_dict = {
    "gradcam": GradCAM,
    "GradCAM": GradCAM,
    "gc": GradCAM,
    "gradcam++": GradCAMpp,
    "GradCAM++": GradCAMpp,
    "GradCAMpp": GradCAMpp,
    "gc++": GradCAMpp,
    "GuidedGradCAM": GuidedGradCAM,
    "ggc": GuidedGradCAM,
    "saliency": SaliencyMap,
    "SaliencyMap": SaliencyMap,
    "sm": SaliencyMap,
    "guidedbackprop": GuidedBackpropagation,
    "GuidedBackpropagation": GuidedBackpropagation,
    "gb": GuidedBackpropagation,
    "IntegratedGradients": IntegratedGradients,
    "ig":IntegratedGradients,
    "OcclusionSensitivity": OcclusionSensitivity,
    "os": OcclusionSensitivity,
    "LimePro": LimePro,
    "lp": LimePro,
    "LimeCon": LimeCon,
    "lc": LimeCon
}
""" Dictionary of implemented XAI Methods in AUCMEDI.

    A key (str) or an initialized XAI Method can be passed to the [aucmedi.xai.decoder.xai_decoder][] function as method parameter.

    ???+ example "Example"
        ```python
        # Select desired XAI Methods
        xai_list = ["gradcam", "gc++", OcclusionSensitivity(model), xai_dict["LimePro"](model), "lc"]

        # Iterate over each method
        for m in xai_list:
            # Compute XAI heatmaps with method m
            heatmaps = xai_decoder(datagen, model, preds, method=m)
        ```

    XAI Methods are based on the abstract base class [aucmedi.xai.methods.xai_base][].
"""



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import os
import tempfile
from tensorflow.keras.callbacks import ModelCheckpoint, CSVLogger
from pathos.helpers import mp   # instead of 'import multiprocessing as mp'
import numpy as np
import shutil
# Internal libraries
from aucmedi import DataGenerator, NeuralNetwork
from aucmedi.sampling import sampling_split, sampling_kfold
from aucmedi.ensemble.aggregate import aggregate_dict
from aucmedi.ensemble.metalearner import metalearner_dict
from aucmedi.ensemble.metalearner.ml_base import Metalearner_Base
from aucmedi.ensemble.aggregate.agg_base import Aggregate_Base

#-----------------------------------------------------#
#            Ensemble Learning: Composite             #
#-----------------------------------------------------#
class Composite:
    """ A Composite class providing functionality for cross-validation and metalearner based ensemble learning.

    The Composite strategy combines the homogeneous [Bagging][aucmedi.ensemble.Bagging] and the heterogeneous
    [Stacking][aucmedi.ensemble.Stacking] technique.

    If a metalearner is selected, a percentage sampling split is applied. For an aggregate function, this is not done.
    The remaining training data is sampled via a cross-validation. For each fold, a different model is trained
    returning into a heterogenous ensemble.
    Predictions for this heterogenous ensemble are combined with the fitted metalearner model or an aggregate function.

    Instead of utilizing the fixed parameters of the [DataGenerator][aucmedi.data_processing.data_generator],
    default paramters for Resizing and Standardize of the associated models are used (if `fixed_datagenerator=True`).

    ???+ example
        ```python
        # Initialize some NeuralNetwork models
        model_a = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet50")
        model_b = NeuralNetwork(n_labels=4, channels=3, architecture="2D.MobileNetV2")
        model_c = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB1")

        # Initialize Composite object
        el = Composite(model_list=[model_a, model_b, model_c],
                       metalearner="logistic_regression", k_fold=3)

        # Initialize training DataGenerator for complete training data
        datagen = DataGenerator(samples_train, "images_dir/",
                                labels=train_labels_ohe, batch_size=3,
                                resize=None, standardize_mode=None)
        # Train neural network and metalearner models
        el.train(datagen, epochs=100)

        # Initialize testing DataGenerator for testing data
        test_gen = DataGenerator(samples_test, "images_dir/",
                                 resize=None, standardize_mode=None)
        # Run Inference
        preds = el.predict(test_gen)
        ```

    !!! warning "Training Time Increase"
        Composite sequentially performs fitting processes for multiple models, which will drastically increase training time.

    ??? warning "DataGenerator re-initialization"
        The passed DataGenerator for the train() and predict() function of the Composite class will be re-initialized!

        This can result in redundant image preparation if `prepare_images=True`.

        Furthermore, the parameters `resize` and `standardize_mode` are automatically re-initialized with
        NeuralNetwork model specific values (`model.meta_standardize` for `standardize_mode` and
        `model.meta_input` for `input_shape`).

        If desired (but not recommended!), it is possible to modify the meta variables of the NeuralNetwork model as follows:
        ```python
        # For input_shape
        model_a = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet50",
                                input_shape=(64,64))
        # For standardize_mode
        model_b = NeuralNetwork(n_labels=4, channels=3, architecture="2D.MobileNetV2")
        model_b.meta_standardize = "torch"
        ```

    ??? warning "NeuralNetwork re-initialization"
        The passed NeuralNetwork for the train() and predict() function of the Composite class will be re-initialized!

        Attention: Metrics are not passed to the processes due to pickling issues.

    ??? info "Technical Details"
        For the training and inference process, each model will create an individual process via the Python multiprocessing package.

        This is crucial as TensorFlow does not fully support the VRAM memory garbage collection in GPUs,
        which is why more and more redundant data pile up with an increasing number of models.

        Via separate processes, it is possible to clean up the TensorFlow environment and rebuild it again for the next model.
    """
    def __init__(self, model_list, metalearner="logistic_regression",
                 k_fold=3, sampling=[0.85, 0.15], fixed_datagenerator=False):
        """ Initialization function for creating a Composite object.

        Args:
            model_list (list of NeuralNetwork):         List of instances of AUCMEDI neural network class.
                                                        The number of models (`len(model_list)`) have to be equal to `k_fold`.
            metalearner (str, Metalearner or Aggregate):Metalearner class instance / a string for an AUCMEDI Metalearner,
                                                        or Aggregate function / a string for an AUCMEDI Aggregate function.
            k_fold (int):                               Number of folds (k) for the Cross-Validation. Must be at least 2.
            sampling (list of float):                   List of percentage values with split sizes. Should be 2x percentage values
                                                        for heterogenous metalearner (must sum up to 1.0).
            fixed_datagenerator (bool):                 Boolean, whether using fixed parameters of passed DataGenerator or
                                                        using default architecture paramters for Resizing and Standardize.
        """
        # Cache class variables
        self.model_list = model_list
        self.metalearner = metalearner
        self.sampling = sampling
        self.k_fold = k_fold
        self.fixed_datagenerator = fixed_datagenerator
        self.sampling_seed = 0
        self.cache_dir = None

        # Initialize Metalearner
        if isinstance(metalearner, str) and metalearner in metalearner_dict:
            self.ml_model = metalearner_dict[metalearner]()
        elif isinstance(metalearner, str) and metalearner in aggregate_dict:
            self.ml_model = aggregate_dict[metalearner]()
        elif isinstance(metalearner, Metalearner_Base) or \
             isinstance(metalearner, Aggregate_Base):
            self.ml_model = metalearner
        else : raise TypeError("Unknown type of Metalearner (neither known " + \
                               "ensembler nor Aggregate or Metalearner class)!")

        # Verify model list length
        if k_fold != len(model_list):
            raise ValueError("Length of model_list and k_fold has to be equal!")

        # Set multiprocessing method to spawn
        mp.set_start_method("spawn", force=True)

    def train(self, training_generator, epochs=20, iterations=None,
              callbacks=[], class_weights=None, transfer_learning=False,
              metalearner_fitting=True):
        """ Training function for fitting the provided NeuralNetwork models.

        The training data will be sampled according to a percentage split in which
        [DataGenerators][aucmedi.data_processing.data_generator.DataGenerator] for model training
        and metalearner training if a metalearner is provided. Else all data is used as model
        training subset. The model training subset is furthermore sampled via cross-validation.

        It is also possible to pass custom Callback classes in order to obtain more information.

        For more information on the fitting process, check out [NeuralNetwork.train()][aucmedi.neural_network.model.NeuralNetwork.train].

        Args:
            training_generator (DataGenerator):     A data generator which will be used for training (will be split according
                                                    to percentage split and k-fold cross-validation sampling).
            epochs (int):                           Number of epochs. A single epoch is defined as one iteration through
                                                    the complete data set.
            iterations (int):                       Number of iterations (batches) in a single epoch.
            callbacks (list of Callback classes):   A list of Callback classes for custom evaluation.
            class_weights (dictionary or list):     A list or dictionary of float values to handle class unbalance.
            transfer_learning (bool):               Option whether a transfer learning training should be performed.
            metalearner_fitting (bool):             Option whether the Metalearner fitting process should be included in the
                                                    Composite training process. The `train_metalearner()` function can also be
                                                    run manually (or repeatedly).
        Returns:
            history (dict):                         A history dictionary from a Keras history object which contains several logs.
        """
        temp_dg = training_generator    # Template DataGenerator variable for faster access
        history_composite = {}           # Final history dictionary

        # Create temporary model directory
        self.cache_dir = tempfile.TemporaryDirectory(prefix="aucmedi.tmp.",
                                                     suffix=".composite")

        # Obtain training data
        x = training_generator.samples
        y = training_generator.labels
        m = training_generator.metadata

        # Apply percentage split sampling for metalearner
        if isinstance(self.ml_model, Metalearner_Base):
            ps_sampling = sampling_split(x, y, m, sampling=self.sampling,
                                         stratified=True, iterative=True,
                                         seed=self.sampling_seed)
            # Pack data according to sampling
            if len(ps_sampling[0]) == 3 : x, y, m = ps_sampling[0]
            else : x, y = ps_sampling[0]

        # Apply cross-validaton sampling
        cv_sampling = sampling_kfold(x, y, m, n_splits=self.k_fold,
                                     stratified=True, iterative=True)

        # Sequentially iterate over model list
        for i in range(len(self.model_list)):
            # Pack data into a tuple
            fold = cv_sampling[i]
            if len(fold) == 4:
                (train_x, train_y, test_x, test_y) = fold
                data = (train_x, train_y, None, test_x, test_y, None)
            else : data = fold

            # Create model specific callback list
            callbacks_model = callbacks.copy()
            # Extend Callback list
            path_model = os.path.join(self.cache_dir.name,
                                      "cv_" + str(i) + ".model.keras")
            cb_mc = ModelCheckpoint(path_model,
                                    monitor="val_loss", verbose=1,
                                    save_best_only=True, mode="min")
            cb_cl = CSVLogger(os.path.join(self.cache_dir.name,
                                                 "cv_" + str(i) + \
                                                 ".logs.csv"),
                              separator=',', append=True)
            callbacks_model.extend([cb_mc, cb_cl])

            # Gather NeuralNetwork parameters
            model_paras = {
                "n_labels": self.model_list[i].n_labels,
                "channels": self.model_list[i].channels,
                "input_shape": self.model_list[i].input_shape,
                "architecture": self.model_list[i].architecture,
                "pretrained_weights": self.model_list[i].pretrained_weights,
                "loss": self.model_list[i].loss,
                "metrics": None,
                "activation_output": self.model_list[i].activation_output,
                "fcl_dropout": self.model_list[i].fcl_dropout,
                "meta_variables": self.model_list[i].meta_variables,
                "learning_rate": self.model_list[i].learning_rate,
            }

            # Gather DataGenerator parameters
            datagen_paras = {"path_imagedir": temp_dg.path_imagedir,
                             "batch_size": temp_dg.batch_size,
                             "data_aug": temp_dg.data_aug,
                             "seed": temp_dg.seed,
                             "subfunctions": temp_dg.subfunctions,
                             "shuffle": temp_dg.shuffle,
                             "standardize_mode": self.model_list[i].meta_standardize,
                             "resize": self.model_list[i].meta_input,
                             "grayscale": temp_dg.grayscale,
                             "prepare_images": temp_dg.prepare_images,
                             "sample_weights": temp_dg.sample_weights,
                             "image_format": temp_dg.image_format,
                             "loader": temp_dg.sample_loader,
                             "workers": temp_dg.workers,
                             "kwargs": temp_dg.kwargs
            }

            # Gather training parameters
            parameters_training = {"epochs": epochs,
                                "iterations": iterations,
                                "callbacks": callbacks_model,
                                "class_weights": class_weights,
                                "transfer_learning": transfer_learning
            }

            # Start training process
            process_queue = mp.Queue()
            process_train = mp.Process(target=__training_process__,
                                       args=(process_queue,
                                             data,
                                             model_paras,
                                             datagen_paras,
                                             parameters_training))
            process_train.start()
            process_train.join()
            cv_history = process_queue.get()
            # Combine logged history objects
            hnn = {"cv_" + str(i) + "." + k: v for k, v in cv_history.items()}
            history_composite = {**history_composite, **hnn}

        # Perform metalearner model training
        if isinstance(self.ml_model, Metalearner_Base):
            self.train_metalearner(temp_dg)

        # Return Composite history object
        return history_composite

    def train_metalearner(self, training_generator):
        """ Training function for fitting the Metalearner model.

        Function will be called automatically in the `train()` function if
        the parameter `metalearner_fitting` is true.

        However, this function can also be called multiple times for training
        different Metalearner types without the need of time-extensive
        re-training of the [NeuralNetwork][aucmedi.neural_network.model] models.

        Args:
            training_generator (DataGenerator):     A data generator which will be used for training (will be split according
                                                    to percentage split).
        """
        # Skipping metalearner training if aggregate function
        if isinstance(self.ml_model, Aggregate_Base) : return

        temp_dg = training_generator    # Template DataGenerator variable for faster access
        preds_ensemble = []

        # Obtain training data
        x = training_generator.samples
        y = training_generator.labels
        m = training_generator.metadata

        # Apply percentage split sampling for metalearner
        if isinstance(self.ml_model, Metalearner_Base):
            ps_sampling = sampling_split(x, y, m, sampling=self.sampling,
                                         stratified=True, iterative=True,
                                         seed=self.sampling_seed)
        # Pack data according to sampling
        if len(ps_sampling[0]) == 3 : data_ensemble = ps_sampling[1]
        else : data_ensemble = (*ps_sampling[1], None)

        # Identify path to model directory
        if isinstance(self.cache_dir, tempfile.TemporaryDirectory):
            path_model_dir = self.cache_dir.name
        else : path_model_dir = self.cache_dir

        # Sequentially iterate over model list
        for i in range(len(self.model_list)):
            # Load current model
            path_model = os.path.join(path_model_dir,
                                      "cv_" + str(i) + ".model.keras")

            # Gather NeuralNetwork parameters
            model_paras = {
                "n_labels": self.model_list[i].n_labels,
                "channels": self.model_list[i].channels,
                "input_shape": self.model_list[i].input_shape,
                "architecture": self.model_list[i].architecture,
                "pretrained_weights": self.model_list[i].pretrained_weights,
                "loss": self.model_list[i].loss,
                "metrics": None,
                "activation_output": self.model_list[i].activation_output,
                "fcl_dropout": self.model_list[i].fcl_dropout,
                "meta_variables": self.model_list[i].meta_variables,
                "learning_rate": self.model_list[i].learning_rate,
            }

            # Gather DataGenerator parameters
            datagen_paras = {"path_imagedir": temp_dg.path_imagedir,
                             "batch_size": temp_dg.batch_size,
                             "data_aug": temp_dg.data_aug,
                             "seed": temp_dg.seed,
                             "subfunctions": temp_dg.subfunctions,
                             "shuffle": temp_dg.shuffle,
                             "standardize_mode": self.model_list[i].meta_standardize,
                             "resize": self.model_list[i].meta_input,
                             "grayscale": temp_dg.grayscale,
                             "prepare_images": temp_dg.prepare_images,
                             "sample_weights": temp_dg.sample_weights,
                             "image_format": temp_dg.image_format,
                             "loader": temp_dg.sample_loader,
                             "workers": temp_dg.workers,
                             "kwargs": temp_dg.kwargs
            }

            # Start inference process for model i
            process_queue = mp.Queue()
            process_pred = mp.Process(target=__prediction_process__,
                                      args=(process_queue,
                                            model_paras,
                                            path_model,
                                            data_ensemble,
                                            datagen_paras))
            process_pred.start()
            process_pred.join()
            preds = process_queue.get()

            # Append preds to ensemble
            preds_ensemble.append(preds)

        # Preprocess prediction ensemble
        preds_ensemble = np.array(preds_ensemble)
        preds_ensemble = np.swapaxes(preds_ensemble, 0, 1)
        s, m, c = preds_ensemble.shape
        x_stack = np.reshape(preds_ensemble, (s, m*c))

        # Start training of stacked metalearner
        if isinstance(self.ml_model, Metalearner_Base):
            (_, y_stack, _) = data_ensemble
            self.ml_model.train(x_stack, y_stack)
            # Store metalearner model to disk
            path_metalearner = os.path.join(path_model_dir,
                                            "metalearner.model.pickle")
            self.ml_model.dump(path_metalearner)

    def predict(self, prediction_generator, return_ensemble=False):
        """ Prediction function for Composite.

        The fitted models and selected Metalearner/Aggregate function will predict classifications
        for the provided [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

        !!! info
            More about Metalearners can be found here: [Metelearner][aucmedi.ensemble.metalearner]

            More about Aggregate functions can be found here: [aggregate][aucmedi.ensemble.aggregate]

        Args:
            prediction_generator (DataGenerator):   A data generator which will be used for inference.
            return_ensemble (bool):                 Option, whether gathered ensemble of predictions should be returned.

        Returns:
            preds (numpy.ndarray):                  A NumPy array of predictions formatted with shape (n_samples, n_labels).
            ensemble (numpy.ndarray):               Optional ensemble of predictions: Will be only passed if `return_ensemble=True`.
                                                    Shape (n_models, n_samples, n_labels).
        """
        # Verify if there is a linked cache dictionary
        con_tmp = (isinstance(self.cache_dir, tempfile.TemporaryDirectory) and \
                   os.path.exists(self.cache_dir.name))
        con_var = (self.cache_dir is not None and \
                   not isinstance(self.cache_dir, tempfile.TemporaryDirectory) \
                   and os.path.exists(self.cache_dir))
        if not con_tmp and not con_var:
            raise FileNotFoundError("Composite instance does not have a valid" \
                                    + "model cache directory!")

        # Initialize some variables
        temp_dg = prediction_generator
        preds_ensemble = []
        preds_final = []

        # Extract data
        data_test = (temp_dg.samples, temp_dg.labels, temp_dg.metadata)

        # Identify path to model directory
        if isinstance(self.cache_dir, tempfile.TemporaryDirectory):
            path_model_dir = self.cache_dir.name
        else : path_model_dir = self.cache_dir

        # Sequentially iterate over model list
        for i in range(len(self.model_list)):
            path_model = os.path.join(path_model_dir,
                                      "cv_" + str(i) + ".model.keras")

            # Gather NeuralNetwork parameters
            model_paras = {
                "n_labels": self.model_list[i].n_labels,
                "channels": self.model_list[i].channels,
                "input_shape": self.model_list[i].input_shape,
                "architecture": self.model_list[i].architecture,
                "pretrained_weights": self.model_list[i].pretrained_weights,
                "loss": self.model_list[i].loss,
                "metrics": None,
                "activation_output": self.model_list[i].activation_output,
                "fcl_dropout": self.model_list[i].fcl_dropout,
                "meta_variables": self.model_list[i].meta_variables,
                "learning_rate": self.model_list[i].learning_rate,
            }

            # Gather DataGenerator parameters
            datagen_paras = {"path_imagedir": temp_dg.path_imagedir,
                             "batch_size": temp_dg.batch_size,
                             "data_aug": temp_dg.data_aug,
                             "seed": temp_dg.seed,
                             "subfunctions": temp_dg.subfunctions,
                             "shuffle": temp_dg.shuffle,
                             "standardize_mode": self.model_list[i].meta_standardize,
                             "resize": self.model_list[i].meta_input,
                             "grayscale": temp_dg.grayscale,
                             "prepare_images": temp_dg.prepare_images,
                             "sample_weights": temp_dg.sample_weights,
                             "image_format": temp_dg.image_format,
                             "loader": temp_dg.sample_loader,
                             "workers": temp_dg.workers,
                             "kwargs": temp_dg.kwargs
            }

            # Start inference process for model i
            process_queue = mp.Queue()
            process_pred = mp.Process(target=__prediction_process__,
                                      args=(process_queue,
                                            model_paras,
                                            path_model,
                                            data_test,
                                            datagen_paras))
            process_pred.start()
            process_pred.join()
            preds = process_queue.get()

            # Append preds to ensemble
            preds_ensemble.append(preds)

        # Preprocess prediction ensemble
        preds_ensemble = np.array(preds_ensemble)
        preds_ensemble = np.swapaxes(preds_ensemble, 0, 1)

        # Apply heterogenous metalearner
        if isinstance(self.ml_model, Metalearner_Base):
            s, m, c = preds_ensemble.shape
            x_stack = np.reshape(preds_ensemble, (s, m*c))
            preds_final = self.ml_model.predict(data=x_stack)
        # Apply homogeneous aggregate function
        elif isinstance(self.ml_model, Aggregate_Base):
            for i in range(preds_ensemble.shape[0]):
                pred_sample = self.ml_model.aggregate(preds_ensemble[i,:,:])
                preds_final.append(pred_sample)

        # Convert prediction list to NumPy
        preds_final = np.asarray(preds_final)

        # Return ensembled predictions
        if return_ensemble : return preds_final, np.swapaxes(preds_ensemble,1,0)
        else : return preds_final

    # Dump model to file
    def dump(self, directory_path):
        """ Store temporary Composite models directory permanently to disk at desired location.

        If the model directory is a provided path which is already persistent on the disk,
        the directory is copied in order to keep original data persistent.

        Args:
            directory_path (str):       Path to store the model directory on disk.
        """
        if self.cache_dir is None:
            raise FileNotFoundError("Composite does not have a valid model cache directory!")
        elif isinstance(self.cache_dir, tempfile.TemporaryDirectory):
            shutil.copytree(self.cache_dir.name, directory_path,
                            dirs_exist_ok=True)
            self.cache_dir.cleanup()
            self.cache_dir = directory_path
        else:
            shutil.copytree(self.cache_dir, directory_path, dirs_exist_ok=True)
            self.cache_dir = directory_path

    # Load model from file
    def load(self, directory_path):
        """ Load a Composite model directory which can be used for Metalearner based inference.

        Args:
            directory_path (str):       Input path, from which the Composite models will be loaded.
        """
        # Check directory existence
        if not os.path.exists(directory_path):
            raise FileNotFoundError("Provided model directory path does not exist!",
                                    directory_path)
        # Check model existence
        for i in range(len(self.model_list)):
            path_model = os.path.join(directory_path,
                                      "cv_" + str(i) + ".model.keras")
            if not os.path.exists(path_model):
                raise FileNotFoundError("Composite model " + str(i) + \
                                        " does not exist!", path_model)
        # If heterogenous metalearner -> load metalearner model file
        if isinstance(self.ml_model, Metalearner_Base):
            path_model = os.path.join(directory_path,
                                      "metalearner.model.pickle")
            if not os.path.exists(path_model):
                raise FileNotFoundError("Metalearner model does not exist!",
                                        path_model)
            self.ml_model.load(path_model)

        # Update model directory
        self.cache_dir = directory_path

#-----------------------------------------------------#
#                     Subroutines                     #
#-----------------------------------------------------#
# Internal function for training a NeuralNetwork model in a separate process
def __training_process__(queue, data, model_paras, datagen_paras, train_paras):
    # Extract data
    (train_x, train_y, train_m, test_x, test_y, test_m) = data
    # Build training DataGenerator
    cv_train_gen = DataGenerator(train_x,
                                 path_imagedir=datagen_paras["path_imagedir"],
                                 labels=train_y,
                                 metadata=train_m,
                                 batch_size=datagen_paras["batch_size"],
                                 data_aug=datagen_paras["data_aug"],
                                 seed=datagen_paras["seed"],
                                 subfunctions=datagen_paras["subfunctions"],
                                 shuffle=datagen_paras["shuffle"],
                                 standardize_mode=datagen_paras["standardize_mode"],
                                 resize=datagen_paras["resize"],
                                 grayscale=datagen_paras["grayscale"],
                                 prepare_images=datagen_paras["prepare_images"],
                                 sample_weights=datagen_paras["sample_weights"],
                                 image_format=datagen_paras["image_format"],
                                 loader=datagen_paras["loader"],
                                 workers=datagen_paras["workers"],
                                 **datagen_paras["kwargs"])
    # Build validation DataGenerator
    cv_val_gen = DataGenerator(test_x,
                               path_imagedir=datagen_paras["path_imagedir"],
                               labels=test_y,
                               metadata=test_m,
                               batch_size=datagen_paras["batch_size"],
                               data_aug=None,
                               seed=datagen_paras["seed"],
                               subfunctions=datagen_paras["subfunctions"],
                               shuffle=False,
                               standardize_mode=datagen_paras["standardize_mode"],
                               resize=datagen_paras["resize"],
                               grayscale=datagen_paras["grayscale"],
                               prepare_images=datagen_paras["prepare_images"],
                               sample_weights=datagen_paras["sample_weights"],
                               image_format=datagen_paras["image_format"],
                               loader=datagen_paras["loader"],
                               workers=datagen_paras["workers"],
                               **datagen_paras["kwargs"])
    # Create NeuralNetwork
    model = NeuralNetwork(**model_paras)
    # Start NeuralNetwork training
    cv_history = model.train(cv_train_gen, cv_val_gen, **train_paras)
    # Store result in cache (which will be returned by the process queue)
    queue.put(cv_history)

# Internal function for inference with a fitted NeuralNetwork model in a separate process
def __prediction_process__(queue, model_paras, path_model, data_test,
                           datagen_paras):
    # Extract data
    (test_x, test_y, test_m) = data_test
    # Create inference DataGenerator
    cv_pred_gen = DataGenerator(test_x,
                                path_imagedir=datagen_paras["path_imagedir"],
                                labels=None,
                                metadata=test_m,
                                batch_size=datagen_paras["batch_size"],
                                data_aug=None,
                                seed=datagen_paras["seed"],
                                subfunctions=datagen_paras["subfunctions"],
                                shuffle=False,
                                standardize_mode=datagen_paras["standardize_mode"],
                                resize=datagen_paras["resize"],
                                grayscale=datagen_paras["grayscale"],
                                prepare_images=datagen_paras["prepare_images"],
                                sample_weights=datagen_paras["sample_weights"],
                                image_format=datagen_paras["image_format"],
                                loader=datagen_paras["loader"],
                                workers=datagen_paras["workers"],
                                **datagen_paras["kwargs"])
    # Create NeuralNetwork
    model = NeuralNetwork(**model_paras)
    # Load model weights from disk
    model.load(path_model)
    # Make prediction
    preds = model.predict(cv_pred_gen)
    # Store prediction results in cache (which will be returned by the process queue)
    queue.put(preds)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" The Augmentation classes of AUCMEDI allow creating interfaces to powerful
    augmentation frameworks and easily integrate them into the AUCMEDI pipeline.

An Augmentation class is a preprocessing method, which is randomly applied on each sample
if provided to a [DataGenerator][aucmedi.data_processing.data_generator.DataGenerator].

???+ warning
    Augmentation should only be applied to a **training** DataGenerator!

    For test-time augmentation, [aucmedi.ensemble.augmenting][] should be used.

Data augmentation is a technique that can be used to artificially expand the size
of a training dataset by creating modified versions of images in the dataset.

The point of data augmentation is, that the model will learn meaningful patterns
instead of meaningless characteristics due to a small data set size.

???+ info "Data Augmentation Interfaces"
    | Interface                                                                                | Description                                                           |
    | ---------------------------------------------------------------------------------------- | --------------------------------------------------------------------- |
    | [ImageAugmentation][aucmedi.data_processing.augmentation.aug_image]                     | Interface to package: Albumentations. Handles only images (2D data).  |
    | [VolumeAugmentation][aucmedi.data_processing.augmentation.aug_volume]                   | Interface to package: Volumentations. Handles only volumes (3D data). |
    | [BatchgeneratorsAugmentation][aucmedi.data_processing.augmentation.aug_batchgenerators] | Interface to package: batchgenerators (DKFZ). Handles images and volumes (2D+3D data). |

**Recommendation:** <br>
- For images (2D data): ImageAugmentation() <br>
- For volumes (3D data): BatchgeneratorsAugmentation() <br>

???+ example
    **For 2D data:**
    ```python
    from aucmedi import *

    aug = ImageAugmentation(flip=True, rotate=True, brightness=True, contrast=True,
                 saturation=True, hue=True, scale=True, crop=False,
                 grid_distortion=False, compression=False, gaussian_noise=False,
                 gaussian_blur=False, downscaling=False, gamma=False,
                 elastic_transform=False)

    datagen = DataGenerator(samples=index_list,
                            path_imagedir="dataset/images/",
                            labels=class_ohe,
                            data_aug=aug,
                            resize=model.meta_input,
                            image_format=image_format)
    ```

    **For 3D data:**
    ```python
    from aucmedi import *

    aug = BatchgeneratorsAugmentation(model.meta_input, mirror=False, rotate=True,
                 scale=True, elastic_transform=False, gaussian_noise=True,
                 brightness=True, contrast=True, gamma=True)

    datagen = DataGenerator(samples=index_list,
                            path_imagedir="dataset/volumes/",
                            labels=class_ohe,
                            data_aug=aug,
                            resize=model.meta_input,
                            image_format=image_format)
    ```
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
from aucmedi.data_processing.augmentation.aug_image import ImageAugmentation
from aucmedi.data_processing.augmentation.aug_volume import VolumeAugmentation
from aucmedi.data_processing.augmentation.aug_batchgenerators import BatchgeneratorsAugmentation



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
import numpy as np
# Internal libraries
from aucmedi import ImageAugmentation, VolumeAugmentation, DataGenerator
from aucmedi.ensemble.aggregate import aggregate_dict
from aucmedi.data_processing.io_loader import image_loader

#-----------------------------------------------------#
#       Ensemble Learning: Inference Augmenting       #
#-----------------------------------------------------#
def predict_augmenting(model, prediction_generator, n_cycles=10, aggregate="mean"):
    """ Inference Augmenting function for automatically augmenting unknown images for prediction.

    The predictions of the augmented images are aggregated via the provided Aggregate function.

    ???+ example
        ```python
        # Import libraries
        from aucmedi.ensemble import predict_augmenting
        from aucmedi import ImageAugmentation, DataGenerator

        # Initialize testing DataGenerator with desired Data Augmentation
        test_aug = ImageAugmentation(flip=True, rotate=True, brightness=False, contrast=False))
        test_gen = DataGenerator(samples_test, "images_dir/",
                                 data_aug=test_aug,
                                 resize=model.meta_input,
                                 standardize_mode=model.meta_standardize)

        # Compute predictions via Augmenting
        preds = predict_augmenting(model, test_gen, n_cycles=15, aggregate="majority_vote")
        ```

    The inclusion of the Aggregate function can be achieved in multiple ways:

    - self-initialization with an AUCMEDI Aggregate function,
    - use a string key to call an AUCMEDI Aggregate function by name, or
    - implementing a custom Aggregate function by extending the [AUCMEDI base class for Aggregate functions][aucmedi.ensemble.aggregate.agg_base]

    !!! info
        Description and list of implemented Aggregate functions can be found here:
        [Aggregate][aucmedi.ensemble.aggregate]

    The Data Augmentation class instance from the DataGenerator will be used for inference augmenting.
    It can either be predefined or remain `None`. If the `data_aug` is `None`, a Data Augmentation class
    instance is automatically created which applies rotation and flipping augmentations.

    ???+ warning
        The passed DataGenerator will be re-initialized!
        This can result in redundant image preparation if `prepare_images=True`.

    ??? reference "Reference for Ensemble Learning Techniques"
        Dominik Müller, Iñaki Soto-Rey and Frank Kramer. (2022).
        An Analysis on Ensemble Learning optimized Medical Image Classification with Deep Convolutional Neural Networks.
        arXiv e-print: [https://arxiv.org/abs/2201.11440](https://arxiv.org/abs/2201.11440)

    Args:
        model (NeuralNetwork):                 Instance of a AUCMEDI neural network class.
        prediction_generator (DataGenerator):   A data generator which will be used for Augmenting based inference.
        n_cycles (int):                         Number of image augmentations, which should be created per sample.
        aggregate (str or aggregate Function):  Aggregate function class instance or a string for an AUCMEDI Aggregate function.
    """
    # Initialize aggregate function if required
    if isinstance(aggregate, str) and aggregate in aggregate_dict:
        agg_fun = aggregate_dict[aggregate]()
    else : agg_fun = aggregate

    # Initialize image augmentation if none provided (only flip, rotate)
    if prediction_generator.data_aug is None and len(model.input_shape) == 3:
        data_aug = ImageAugmentation(flip=True, rotate=True, scale=False,
                                     brightness=False, contrast=False,
                                     saturation=False, hue=False, crop=False,
                                     grid_distortion=False, compression=False,
                                     gamma=False, gaussian_noise=False,
                                     gaussian_blur=False, downscaling=False,
                                     elastic_transform=False)
    elif prediction_generator.data_aug is None and len(model.input_shape) == 4:
        data_aug = VolumeAugmentation(flip=True, rotate=True, scale=False,
                                      brightness=False, contrast=False,
                                      saturation=False, hue=False, crop=False,
                                      grid_distortion=False, compression=False,
                                      gamma=False, gaussian_noise=False,
                                      gaussian_blur=False, downscaling=False,
                                      elastic_transform=False)
    else : data_aug = prediction_generator.data_aug
    # Multiply sample list for prediction according to number of cycles
    samples_aug = np.repeat(prediction_generator.samples, n_cycles)

    # Re-initialize DataGenerator for inference
    aug_gen = DataGenerator(samples_aug,
                            path_imagedir=prediction_generator.path_imagedir,
                            labels=None,
                            metadata=prediction_generator.metadata,
                            batch_size=prediction_generator.batch_size,
                            data_aug=data_aug,
                            seed=prediction_generator.seed,
                            subfunctions=prediction_generator.subfunctions,
                            shuffle=False,
                            standardize_mode=prediction_generator.standardize_mode,
                            resize=prediction_generator.resize,
                            grayscale=prediction_generator.grayscale,
                            prepare_images=prediction_generator.prepare_images,
                            sample_weights=None,
                            image_format=prediction_generator.image_format,
                            loader=prediction_generator.sample_loader,
                            workers=prediction_generator.workers,
                            **prediction_generator.kwargs)

    # Compute predictions with provided model
    preds_all = model.predict(aug_gen)

    # Ensemble inferences via aggregate function
    preds_ensembled = []
    for i in range(0, len(prediction_generator.samples)):
        # Identify subset for a single sample
        j = i*n_cycles
        subset = preds_all[j:j+n_cycles]
        # Aggregate predictions
        pred_sample = agg_fun.aggregate(subset)
        # Add prediction to prediction list
        preds_ensembled.append(pred_sample)
    # Convert prediction list to NumPy
    preds_ensembled = np.asarray(preds_ensembled)

    # Return ensembled predictions
    return preds_ensembled



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
import random
#Internal libraries
from aucmedi import ImageAugmentation, VolumeAugmentation, BatchgeneratorsAugmentation

#-----------------------------------------------------#
#             Unittest: Image Augmentation            #
#-----------------------------------------------------#
class AugmentationTEST(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        random.seed(1234)
        np.random.seed(1234)
        # Create Grayscale data
        img_gray_2D = np.random.rand(16, 16, 1) * 254
        self.imgGRAY2d = np.float32(img_gray_2D)
        img_gray_3D = np.random.rand(16, 16, 16, 1) * 254
        self.imgGRAY3d = np.float32(img_gray_3D)
        # Create RGB data
        img_rgb_2D = np.random.rand(16, 16, 3) * 254
        self.imgRGB2d = np.float32(img_rgb_2D)
        img_rgb_3D = np.random.rand(16, 16, 16, 3) * 254
        self.imgRGB3d = np.float32(img_rgb_3D)

    #-------------------------------------------------#
    #               Image Functionality               #
    #-------------------------------------------------#
    # Class Creation
    def test_IMAGE_create(self):
        data_aug = ImageAugmentation()
        self.assertIsInstance(data_aug, ImageAugmentation)

    # Application
    def test_IMAGE_application(self):
        data_aug = ImageAugmentation(flip=True, rotate=True,
                     brightness=True, contrast=True, saturation=True,
                     hue=True, scale=True, crop=True, grid_distortion=True,
                     compression=True, gaussian_noise=True,
                     gaussian_blur=True, downscaling=True, gamma=True,
                     elastic_transform=True)
        data_aug.aug_elasticTransform_p = 1.0
        data_aug.aug_gamma_p = 1.0
        data_aug.aug_downscaling_p = 1.0
        data_aug.aug_gaussianBlur_p = 1.0
        data_aug.aug_gaussianNoise_p = 1.0
        data_aug.aug_compression_p = 1.0
        data_aug.aug_gridDistortion_p = 1.0
        data_aug.aug_crop_p = 1.0
        data_aug.aug_scale_p = 1.0
        data_aug.aug_hue_p = 1.0
        data_aug.aug_saturation_p = 1.0
        data_aug.aug_contrast_p = 1.0
        data_aug.aug_brightness_p = 1.0
        data_aug.aug_rotate_p = 1.0
        data_aug.aug_flip_p = 1.0
        data_aug.aug_crop_shape = (8, 8)
        data_aug.build()
        data_augGRAY = data_aug.apply(self.imgGRAY2d)
        data_augRGB = data_aug.apply(self.imgRGB2d)
        self.assertFalse(np.array_equal(data_augGRAY, self.imgGRAY2d))
        self.assertFalse(np.array_equal(data_augRGB, self.imgRGB2d))

    # Rebuild Augmentation Operator
    def test_IMAGE_rebuild(self):
        data_aug = ImageAugmentation(flip=False, rotate=False,
                     brightness=False, contrast=False, saturation=False,
                     hue=False, scale=False, crop=False, grid_distortion=False,
                     compression=False, gaussian_noise=False,
                     gaussian_blur=False, downscaling=False, gamma=False,
                     elastic_transform=False)
        data_augRGB = data_aug.apply(self.imgRGB2d)
        self.assertTrue(np.array_equal(data_augRGB, self.imgRGB2d))
        data_aug.aug_flip = True
        data_aug.aug_flip_p = 1.0
        data_aug.build()
        data_augRGB = data_aug.apply(self.imgRGB2d)
        self.assertFalse(np.array_equal(data_augRGB, self.imgRGB2d))

    #-------------------------------------------------#
    #               Volume Functionality              #
    #-------------------------------------------------#
    # Class Creation
    def test_VOLUME_create(self):
        data_aug = VolumeAugmentation()
        self.assertIsInstance(data_aug, VolumeAugmentation)

    # Application
    def test_VOLUME_application(self):
        data_aug = VolumeAugmentation(flip=True, rotate=True,
                     brightness=True, contrast=True, saturation=True,
                     hue=True, scale=True, crop=True, grid_distortion=True,
                     compression=True, gaussian_noise=True,
                     gaussian_blur=True, downscaling=True, gamma=True,
                     elastic_transform=True)
        data_aug.aug_elasticTransform_p = 1.0
        data_aug.aug_gamma_p = 1.0
        data_aug.aug_downscaling_p = 1.0
        data_aug.aug_gaussianBlur_p = 1.0
        data_aug.aug_gaussianNoise_p = 1.0
        data_aug.aug_compression_p = 1.0
        data_aug.aug_gridDistortion_p = 1.0
        data_aug.aug_crop_p = 1.0
        data_aug.aug_scale_p = 1.0
        data_aug.aug_hue_p = 1.0
        data_aug.aug_saturation_p = 1.0
        data_aug.aug_contrast_p = 1.0
        data_aug.aug_brightness_p = 1.0
        data_aug.aug_rotate_p = 1.0
        data_aug.aug_flip_p = 1.0
        data_aug.aug_crop_shape = (8, 8, 8)
        data_aug.build()
        data_augGRAY = data_aug.apply(self.imgGRAY3d)
        data_augRGB = data_aug.apply(self.imgRGB3d)
        self.assertFalse(np.array_equal(data_augGRAY, self.imgGRAY3d))
        self.assertFalse(np.array_equal(data_augRGB, self.imgRGB3d))

    # Rebuild Augmentation Operator
    def test_VOLUME_rebuild(self):
        data_aug = VolumeAugmentation(flip=False, rotate=False,
                     brightness=False, contrast=False, saturation=False,
                     hue=False, scale=False, crop=False, grid_distortion=False,
                     compression=False, gaussian_noise=False,
                     gaussian_blur=False, downscaling=False, gamma=False,
                     elastic_transform=False)
        data_augRGB = data_aug.apply(self.imgRGB3d)
        self.assertTrue(np.array_equal(data_augRGB, self.imgRGB3d))
        data_aug.aug_flip = True
        data_aug.aug_flip_p = 1.0
        data_aug.build()
        data_augRGB = data_aug.apply(self.imgRGB3d)
        self.assertFalse(np.array_equal(data_augRGB, self.imgRGB3d))

    #-------------------------------------------------#
    #          Batchgenerators Functionality          #
    #-------------------------------------------------#
    # Class Creation
    def test_BATCHGENERATORS_create(self):
        data_aug = BatchgeneratorsAugmentation(image_shape=(16,16,16))
        self.assertIsInstance(data_aug, BatchgeneratorsAugmentation)

    # Application
    def test_BATCHGENERATORS_application(self):
        data_aug = BatchgeneratorsAugmentation(image_shape=(16,16,16),
                        mirror=True, rotate=True, scale=True,
                        elastic_transform=True, gaussian_noise=True,
                        brightness=True, contrast=True, gamma=True)
        data_aug.aug_elasticTransform_p = 1.0
        data_aug.aug_gamma_p = 1.0
        data_aug.aug_gaussianNoise_p = 1.0
        data_aug.aug_scale_p = 1.0
        data_aug.aug_contrast_p = 1.0
        data_aug.aug_brightness_p = 1.0
        data_aug.aug_rotate_p = 1.0
        data_aug.aug_mirror_p = 1.0
        data_aug.build()
        data_augGRAY = data_aug.apply(self.imgGRAY3d)
        data_augRGB = data_aug.apply(self.imgRGB3d)
        self.assertFalse(np.array_equal(data_augGRAY, self.imgGRAY3d))
        self.assertFalse(np.array_equal(data_augRGB, self.imgRGB3d))

    # Rebuild Augmentation Operator
    def test_BATCHGENERATORS_rebuild(self):
        data_aug = BatchgeneratorsAugmentation(image_shape=(16,16,16),
                        mirror=False, rotate=False, scale=False,
                        elastic_transform=False, gaussian_noise=False,
                        brightness=False, contrast=False, gamma=False)
        data_augRGB = data_aug.apply(self.imgRGB3d)
        self.assertTrue(np.array_equal(data_augRGB, self.imgRGB3d))
        data_aug.aug_mirror = True
        data_aug.aug_mirror_p = 1.0
        data_aug.build()
        data_augRGB = data_aug.apply(self.imgRGB3d)
        self.assertFalse(np.array_equal(data_augRGB, self.imgRGB3d))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import numpy as np
from PIL import Image
import tempfile
import os
from tensorflow.keras.callbacks import CSVLogger
#Internal libraries
from aucmedi.utils.callbacks import *
from aucmedi import *

#-----------------------------------------------------#
#                  Unittest: Utility                  #
#-----------------------------------------------------#
class UtilityTEST(unittest.TestCase):
    # Create random imaging data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")

        # Create RGB data
        self.sampleList_rgb = []
        for i in range(0, 1):
            img_rgb = np.random.rand(32, 32, 3) * 255
            imgRGB_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "image.sample_" + str(i) + ".RGB.png"
            path_sampleRGB = os.path.join(self.tmp_data.name, index)
            imgRGB_pillow.save(path_sampleRGB)
            self.sampleList_rgb.append(index)

        # Create classification labels
        self.labels_ohe = np.zeros((1, 4), dtype=np.uint8)
        for i in range(0, 1):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1

        # Create RGB Data Generator
        self.datagen = DataGenerator(self.sampleList_rgb,
                                     self.tmp_data.name,
                                     labels=self.labels_ohe,
                                     #resize=(32, 32),
                                     grayscale=False, batch_size=1)

    #-------------------------------------------------#
    #             Callbacks: CSV2history              #
    #-------------------------------------------------#
    def test_Callbacks_csv2history(self):
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".model")
        path_csv = os.path.join(self.tmp_data.name, "testing.csv")
        csvlog = CSVLogger(path_csv)

        model = NeuralNetwork(n_labels=4, channels=3)
        hist_returned = model.train(training_generator=self.datagen,
                                    validation_generator=self.datagen,
                                    epochs=3, callbacks=[csvlog])

        hist_loaded = csv_to_history(path_csv)
        del hist_loaded["epoch"]

        for key in hist_returned:
            self.assertTrue(key in hist_returned and key in hist_loaded)
            self.assertTrue(len(hist_returned[key]) == len(hist_loaded[key]))



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
#External libraries
import unittest
import tempfile
import os
from PIL import Image
import numpy as np
#Internal libraries
from aucmedi.neural_network.architectures.image import *
from aucmedi.neural_network.architectures import supported_standardize_mode as sdm_global
from aucmedi.neural_network.architectures import Classifier, architecture_dict
from aucmedi import *
from aucmedi.data_processing.subfunctions import Resize

#-----------------------------------------------------#
#               Unittest: Architectures               #
#-----------------------------------------------------#
class ArchitecturesImageTEST(unittest.TestCase):
    # Create random imaging and classification data
    @classmethod
    def setUpClass(self):
        np.random.seed(1234)
        # Initialize temporary directory
        self.tmp_data = tempfile.TemporaryDirectory(prefix="tmp.aucmedi.",
                                                    suffix=".data")
        # Create Grayscale data
        self.sampleList_gray = []
        for i in range(0, 1):
            img_gray = np.random.rand(32, 32) * 255
            imgGRAY_pillow = Image.fromarray(img_gray.astype(np.uint8))
            index = "image.sample_" + str(i) + ".GRAY.png"
            path_sampleGRAY = os.path.join(self.tmp_data.name, index)
            imgGRAY_pillow.save(path_sampleGRAY)
            self.sampleList_gray.append(index)

        # Create RGB data
        self.sampleList_rgb = []
        for i in range(0, 1):
            img_rgb = np.random.rand(32, 32, 3) * 255
            imgRGB_pillow = Image.fromarray(img_rgb.astype(np.uint8))
            index = "image.sample_" + str(i) + ".RGB.png"
            path_sampleRGB = os.path.join(self.tmp_data.name, index)
            imgRGB_pillow.save(path_sampleRGB)
            self.sampleList_rgb.append(index)

        # Create classification labels
        self.labels_ohe = np.zeros((1, 4), dtype=np.uint8)
        for i in range(0, 1):
            class_index = np.random.randint(0, 4)
            self.labels_ohe[i][class_index] = 1

        # Create Grayscale Data Generator
        self.datagen_GRAY = DataGenerator(self.sampleList_gray,
                                          self.tmp_data.name,
                                          labels=self.labels_ohe,
                                          resize=(32, 32),
                                          grayscale=True, batch_size=1)
        # Create RGB Data Generator
        self.datagen_RGB = DataGenerator(self.sampleList_rgb,
                                         self.tmp_data.name,
                                         labels=self.labels_ohe,
                                         resize=(32, 32),
                                         grayscale=False, batch_size=1)

    #-------------------------------------------------#
    #              Architecture: Vanilla              #
    #-------------------------------------------------#
    def test_Vanilla(self):
        arch = Vanilla(Classifier(n_labels=4), channels=1,
                                    input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = Vanilla(Classifier(n_labels=4), channels=3,
                                    input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.Vanilla",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["Vanilla"] == "z-score")
        self.assertTrue(sdm_global["2D.Vanilla"] == "z-score")

    #-------------------------------------------------#
    #              Architecture: ResNet50             #
    #-------------------------------------------------#
    def test_ResNet50(self):
        arch = ResNet50(Classifier(n_labels=4), channels=1,
                                     input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ResNet50(Classifier(n_labels=4), channels=3,
                                     input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet50",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ResNet50"] == "caffe")
        self.assertTrue(sdm_global["2D.ResNet50"] == "caffe")

    #-------------------------------------------------#
    #             Architecture: ResNet101             #
    #-------------------------------------------------#
    def test_ResNet101(self):
        arch = ResNet101(Classifier(n_labels=4), channels=1,
                                      input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ResNet101(Classifier(n_labels=4), channels=3,
                                      input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet101",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ResNet101"] == "caffe")
        self.assertTrue(sdm_global["2D.ResNet101"] == "caffe")

    #-------------------------------------------------#
    #             Architecture: ResNet152             #
    #-------------------------------------------------#
    def test_ResNet152(self):
        arch = ResNet152(Classifier(n_labels=4), channels=1,
                                      input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ResNet152(Classifier(n_labels=4), channels=3,
                                      input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet152",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ResNet152"] == "caffe")
        self.assertTrue(sdm_global["2D.ResNet152"] == "caffe")

    #-------------------------------------------------#
    #             Architecture: ResNet50V2            #
    #-------------------------------------------------#
    def test_ResNet50V2(self):
        arch = ResNet50V2(Classifier(n_labels=4), channels=1,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ResNet50V2(Classifier(n_labels=4), channels=3,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet50V2",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ResNet50V2"] == "tf")
        self.assertTrue(sdm_global["2D.ResNet50V2"] == "tf")

    #-------------------------------------------------#
    #             Architecture: ResNet101V2           #
    #-------------------------------------------------#
    def test_ResNet101V2(self):
        arch = ResNet101V2(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ResNet101V2(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet101V2",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ResNet101V2"] == "tf")
        self.assertTrue(sdm_global["2D.ResNet101V2"] == "tf")

    #-------------------------------------------------#
    #             Architecture: ResNet152V2           #
    #-------------------------------------------------#
    def test_ResNet152V2(self):
        arch = ResNet152V2(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ResNet152V2(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ResNet152V2",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ResNet152V2"] == "tf")
        self.assertTrue(sdm_global["2D.ResNet152V2"] == "tf")

    #-------------------------------------------------#
    #            Architecture: DenseNet121            #
    #-------------------------------------------------#
    def test_DenseNet121(self):
        arch = DenseNet121(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = DenseNet121(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.DenseNet121",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["DenseNet121"] == "torch")
        self.assertTrue(sdm_global["2D.DenseNet121"] == "torch")

    #-------------------------------------------------#
    #            Architecture: DenseNet169            #
    #-------------------------------------------------#
    def test_DenseNet169(self):
        arch = DenseNet169(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = DenseNet169(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.DenseNet169",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["DenseNet169"] == "torch")
        self.assertTrue(sdm_global["2D.DenseNet169"] == "torch")

    #-------------------------------------------------#
    #            Architecture: DenseNet201            #
    #-------------------------------------------------#
    def test_DenseNet201(self):
        arch = DenseNet201(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = DenseNet201(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.DenseNet201",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["DenseNet201"] == "torch")
        self.assertTrue(sdm_global["2D.DenseNet201"] == "torch")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB0          #
    #-------------------------------------------------#
    def test_EfficientNetB0(self):
        arch = EfficientNetB0(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB0(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB0",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB0"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB0"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB1          #
    #-------------------------------------------------#
    def test_EfficientNetB1(self):
        arch = EfficientNetB1(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB1(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB1",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB1"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB1"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB2          #
    #-------------------------------------------------#
    def test_EfficientNetB2(self):
        arch = EfficientNetB2(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB2(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB2",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB2"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB2"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB3          #
    #-------------------------------------------------#
    def test_EfficientNetB2(self):
        arch = EfficientNetB3(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB3(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB3",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB3"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB3"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB4          #
    #-------------------------------------------------#
    def test_EfficientNetB4(self):
        arch = EfficientNetB4(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB4(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB4",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB4"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB4"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB5          #
    #-------------------------------------------------#
    def test_EfficientNetB5(self):
        arch = EfficientNetB5(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB5(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB5",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB5"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB5"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB6          #
    #-------------------------------------------------#
    def test_EfficientNetB6(self):
        arch = EfficientNetB6(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB6(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB6",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB6"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB6"] == "caffe")

    #-------------------------------------------------#
    #           Architecture: EfficientNetB7          #
    #-------------------------------------------------#
    def test_EfficientNetB7(self):
        arch = EfficientNetB7(Classifier(n_labels=4), channels=1,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = EfficientNetB7(Classifier(n_labels=4), channels=3,
                                           input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.EfficientNetB7",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["EfficientNetB7"] == "caffe")
        self.assertTrue(sdm_global["2D.EfficientNetB7"] == "caffe")

    #-------------------------------------------------#
    #             Architecture: MobileNet             #
    #-------------------------------------------------#
    def test_MobileNet(self):
        arch = MobileNet(Classifier(n_labels=4), channels=1,
                                      input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = MobileNet(Classifier(n_labels=4), channels=3,
                                      input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.MobileNet",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["MobileNet"] == "tf")
        self.assertTrue(sdm_global["2D.MobileNet"] == "tf")

    #-------------------------------------------------#
    #            Architecture: MobileNetV2            #
    #-------------------------------------------------#
    def test_MobileNetV2(self):
        arch = MobileNetV2(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = MobileNetV2(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.MobileNetV2",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["MobileNetV2"] == "tf")
        self.assertTrue(sdm_global["2D.MobileNetV2"] == "tf")

    #-------------------------------------------------#
    #           Architecture: NASNetMobile            #
    #-------------------------------------------------#
    def test_NASNetMobile(self):
        arch = NASNetMobile(Classifier(n_labels=4), channels=1,
                                         input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = NASNetMobile(Classifier(n_labels=4), channels=3,
                                         input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.NASNetMobile",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["NASNetMobile"] == "tf")
        self.assertTrue(sdm_global["2D.NASNetMobile"] == "tf")

    #-------------------------------------------------#
    #            Architecture: NASNetLarge            #
    #-------------------------------------------------#
    def test_NASNetLarge(self):
        arch = NASNetLarge(Classifier(n_labels=4), channels=1,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = NASNetLarge(Classifier(n_labels=4), channels=3,
                                        input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.NASNetLarge",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["NASNetLarge"] == "tf")
        self.assertTrue(sdm_global["2D.NASNetLarge"] == "tf")

    #-------------------------------------------------#
    #         Architecture: InceptionResNetV2         #
    #-------------------------------------------------#
    def test_InceptionResNetV2(self):
        self.datagen_GRAY.sf_resize = Resize(shape=(75, 75))
        self.datagen_RGB.sf_resize = Resize(shape=(75, 75))
        arch = InceptionResNetV2(Classifier(n_labels=4), channels=1,
                                              input_shape=(75, 75))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = InceptionResNetV2(Classifier(n_labels=4), channels=3,
                                              input_shape=(75, 75))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.InceptionResNetV2",
                               input_shape=(75, 75))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["InceptionResNetV2"] == "tf")
        self.assertTrue(sdm_global["2D.InceptionResNetV2"] == "tf")
        self.datagen_GRAY.sf_resize = Resize(shape=(32, 32))
        self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #            Architecture: InceptionV3            #
    #-------------------------------------------------#
    def test_InceptionV3(self):
        self.datagen_GRAY.sf_resize = Resize(shape=(75, 75))
        self.datagen_RGB.sf_resize = Resize(shape=(75, 75))
        arch = InceptionV3(Classifier(n_labels=4), channels=1,
                                        input_shape=(75, 75))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = InceptionV3(Classifier(n_labels=4), channels=3,
                                        input_shape=(75, 75))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.InceptionV3",
                               input_shape=(75, 75))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["InceptionV3"] == "tf")
        self.assertTrue(sdm_global["2D.InceptionV3"] == "tf")
        self.datagen_GRAY.sf_resize = Resize(shape=(32, 32))
        self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #               Architecture: VGG16               #
    #-------------------------------------------------#
    def test_VGG16(self):
        arch = VGG16(Classifier(n_labels=4), channels=1,
                                  input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = VGG16(Classifier(n_labels=4), channels=3,
                                  input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.VGG16",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["VGG16"] == "caffe")
        self.assertTrue(sdm_global["2D.VGG16"] == "caffe")

    #-------------------------------------------------#
    #               Architecture: VGG19               #
    #-------------------------------------------------#
    def test_VGG19(self):
        arch = VGG19(Classifier(n_labels=4), channels=1,
                                  input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = VGG19(Classifier(n_labels=4), channels=3,
                                  input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.VGG19",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["VGG19"] == "caffe")
        self.assertTrue(sdm_global["2D.VGG19"] == "caffe")

    #-------------------------------------------------#
    #              Architecture: Xception             #
    #-------------------------------------------------#
    def test_Xception(self):
        self.datagen_GRAY.sf_resize = Resize(shape=(71, 71))
        self.datagen_RGB.sf_resize = Resize(shape=(71, 71))
        arch = Xception(Classifier(n_labels=4), channels=1,
                                     input_shape=(71, 71))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = Xception(Classifier(n_labels=4), channels=3,
                                     input_shape=(71, 71))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.Xception",
                               input_shape=(71, 71))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["Xception"] == "tf")
        self.assertTrue(sdm_global["2D.Xception"] == "tf")
        self.datagen_GRAY.sf_resize = Resize(shape=(32, 32))
        self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #              Architecture: ViT B16              #
    #-------------------------------------------------#
    # Functionality and Interoperability testing deactived due to too intensive RAM requirements
    # def test_ViT_B16(self):
        # self.datagen_RGB.sf_resize = Resize(shape=(224, 224))
        # arch = ViT_B16(Classifier(n_labels=4), channels=3,
        #                             input_shape=(224, 224))
        # model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        # model.predict(self.datagen_RGB)
        # model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ViT_B16",
        #                        input_shape=(224, 224))
        # try : model.model.summary()
        # except : raise Exception()
        # self.assertTrue(supported_standardize_mode["ViT_B16"] == "tf")
        # self.assertTrue(sdm_global["2D.ViT_B16"] == "tf")
        # self.assertTrue("2D.ViT_B16" in architecture_dict)
        # self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #              Architecture: ViT B32              #
    #-------------------------------------------------#
    # Functionality and Interoperability testing deactived due to too intensive RAM requirements
    # def test_ViT_B32(self):
        # self.datagen_RGB.sf_resize = Resize(shape=(224, 224))
        # arch = ViT_B32(Classifier(n_labels=4), channels=3,
        #                             input_shape=(224, 224))
        # model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        # model.predict(self.datagen_RGB)
        # model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ViT_B32",
        #                        input_shape=(224, 224))
        # try : model.model.summary()
        # except : raise Exception()
        # self.assertTrue(supported_standardize_mode["ViT_B32"] == "tf")
        # self.assertTrue(sdm_global["2D.ViT_B32"] == "tf")
        # self.assertTrue("2D.ViT_B32" in architecture_dict)
        # self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #              Architecture: ViT L16              #
    #-------------------------------------------------#
    # Functionality and Interoperability testing deactived due to too intensive RAM requirements
    # def test_ViT_L16(self):
        # self.datagen_RGB.sf_resize = Resize(shape=(384, 384))
        # arch = ViT_L16(Classifier(n_labels=4), channels=3,
        #                             input_shape=(384, 384))
        # model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        # model.predict(self.datagen_RGB)
        # model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ViT_L16",
        #                        input_shape=(384, 384))
        # try : model.model.summary()
        # except : raise Exception()
        # self.assertTrue(supported_standardize_mode["ViT_L16"] == "tf")
        # self.assertTrue(sdm_global["2D.ViT_L16"] == "tf")
        # self.assertTrue("2D.ViT_L16" in architecture_dict)
        # self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #              Architecture: ViT L32              #
    #-------------------------------------------------#
    # Functionality and Interoperability testing deactived due to too intensive RAM requirements
    # def test_ViT_L32(self):
        # self.datagen_RGB.sf_resize = Resize(shape=(384, 384))
        # arch = ViT_L32(Classifier(n_labels=4), channels=3,
        #                             input_shape=(384, 384))
        # model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        # model.predict(self.datagen_RGB)
        # model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ViT_L32",
        #                        input_shape=(384, 384))
        # try : model.model.summary()
        # except : raise Exception()
        # self.assertTrue(supported_standardize_mode["ViT_L32"] == "tf")
        # self.assertTrue(sdm_global["2D.ViT_L32"] == "tf")
        # self.assertTrue("2D.ViT_L32" in architecture_dict)
        # self.datagen_RGB.sf_resize = Resize(shape=(32, 32))

    #-------------------------------------------------#
    #            Architecture: ConvNeXtBase           #
    #-------------------------------------------------#
    def test_ConvNeXtBase(self):
        arch = ConvNeXtBase(Classifier(n_labels=4), channels=1,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ConvNeXtBase(Classifier(n_labels=4), channels=3,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ConvNeXtBase",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ConvNeXtBase"] == None)
        self.assertTrue(sdm_global["2D.ConvNeXtBase"] == None)

    #-------------------------------------------------#
    #            Architecture: ConvNeXtTiny           #
    #-------------------------------------------------#
    def test_ConvNeXtTiny(self):
        arch = ConvNeXtTiny(Classifier(n_labels=4), channels=1,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ConvNeXtTiny(Classifier(n_labels=4), channels=3,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ConvNeXtTiny",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ConvNeXtTiny"] == None)
        self.assertTrue(sdm_global["2D.ConvNeXtTiny"] == None)

    #-------------------------------------------------#
    #           Architecture: ConvNeXtSmall           #
    #-------------------------------------------------#
    def test_ConvNeXtSmall(self):
        arch = ConvNeXtSmall(Classifier(n_labels=4), channels=1,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ConvNeXtSmall(Classifier(n_labels=4), channels=3,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ConvNeXtSmall",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ConvNeXtSmall"] == None)
        self.assertTrue(sdm_global["2D.ConvNeXtSmall"] == None)

    #-------------------------------------------------#
    #           Architecture: ConvNeXtLarge           #
    #-------------------------------------------------#
    def test_ConvNeXtLarge(self):
        arch = ConvNeXtLarge(Classifier(n_labels=4), channels=1,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=1, architecture=arch)
        model.predict(self.datagen_GRAY)
        arch = ConvNeXtLarge(Classifier(n_labels=4), channels=3,
                                       input_shape=(32, 32))
        model = NeuralNetwork(n_labels=4, channels=3, architecture=arch)
        model.predict(self.datagen_RGB)
        model = NeuralNetwork(n_labels=4, channels=3, architecture="2D.ConvNeXtLarge",
                               input_shape=(32, 32))
        try : model.model.summary()
        except : raise Exception()
        self.assertTrue(supported_standardize_mode["ConvNeXtLarge"] == None)
        self.assertTrue(sdm_global["2D.ConvNeXtLarge"] == None)



#==============================================================================#
#  Author:       Dominik Müller                                                #
#  Copyright:    2024 IT-Infrastructure for Translational Medical Research,    #
#                University of Augsburg                                        #
#                                                                              #
#  This program is free software: you can redistribute it and/or modify        #
#  it under the terms of the GNU General Public License as published by        #
#  the Free Software Foundation, either version 3 of the License, or           #
#  (at your option) any later version.                                         #
#                                                                              #
#  This program is distributed in the hope that it will be useful,             #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#==============================================================================#
#-----------------------------------------------------#
#                    Documentation                    #
#-----------------------------------------------------#
""" The classification variant of the ResNet18 architecture.

| Architecture Variable    | Value                      |
| ------------------------ | -------------------------- |
| Key in architecture_dict | "3D.ResNet18"              |
| Input_shape              | (64, 64, 64)               |
| Standardization          | "grayscale"                |

???+ abstract "Reference - Implementation"
    Solovyev, Roman & Kalinin, Alexandr & Gabruseva, Tatiana. (2021). <br>
    3D Convolutional Neural Networks for Stalled Brain Capillary Detection. <br>
    [https://github.com/ZFTurbo/classification_models_3D](https://github.com/ZFTurbo/classification_models_3D) <br>

???+ abstract "Reference - Publication"
    Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun. 10 Dec 2015.
    Deep Residual Learning for Image Recognition.
    <br>
    [https://arxiv.org/abs/1512.03385](https://arxiv.org/abs/1512.03385)
"""
#-----------------------------------------------------#
#                   Library imports                   #
#-----------------------------------------------------#
# External libraries
from classification_models_3D.tfkeras import Classifiers
# Internal libraries
from aucmedi.neural_network.architectures import Architecture_Base

#-----------------------------------------------------#
#            Architecture class: ResNet18             #
#-----------------------------------------------------#
class ResNet18(Architecture_Base):
    #---------------------------------------------#
    #                Initialization               #
    #---------------------------------------------#
    def __init__(self, classification_head, channels, input_shape=(64, 64, 64),
                 pretrained_weights=False):
        self.classifier = classification_head
        self.input = input_shape + (channels,)
        self.pretrained_weights = pretrained_weights

    #---------------------------------------------#
    #                Create Model                 #
    #---------------------------------------------#
    def create_model(self):
        # Get pretrained image weights from imagenet if desired
        if self.pretrained_weights : model_weights = "imagenet"
        else : model_weights = None

        # Obtain ResNet18 as base model
        BaseModel, preprocess_input = Classifiers.get("resnet18")
        base_model = BaseModel(include_top=False, weights=model_weights,
                               input_tensor=None, input_shape=self.input,
                               pooling=None)
        top_model = base_model.output

        # Add classification head
        model = self.classifier.build(model_input=base_model.input,
                                      model_output=top_model)

        # Return created model
        return model
